Type TCPPWriter Extends TCodeWriter
	Const NEWLINE:String = "~n"
	
	Field CurrentNode:TASTNode
	Field CurrentLine:Int
	Field CurrentScope:TDebugScope
	
	Field IndentationLevel:Int
	Field IndentString:String
	
	Field OutputQueue:TList
	
	Method New()
		OutputQueue:TList = New TList
		CurrentScope = New TDebugScope.Create(Null)
		CurrentLine = 1
	End Method
	
	Method SetCurrent(Source:TASTNode)
		CurrentNode = Source
	End Method
	
	Method AddStandardInclude(File:String)
		AddRaw("#include <" + File + ">")
		AddNewline()
	End Method
	
	Method AddInclude(File:String)
		AddRaw("#include ~q" + File + "~q")
		AddNewline()
	End Method
	
	Method AddNewline()
		Enqueue(NEWLINE)
		CurrentLine :+ 1
	End Method
	
	Method AddLocal(Variable:TVariableDeclNode)
		CurrentScope.AddChild(New TDebugVariable.Create(Variable, CurrentLine))
	End Method
	
	'Method AddFunction
	'Method AddClass
	
	Method AddStatement(Source:TASTNode)
		BeginStatement(Source)
		Source.GenerateCode(Self)
		EndStatement()
	End Method
	
	Method AddStatementRaw(Raw:String)
		AddRaw(IndentString + Raw + ";")
		AddNewline()
	End Method
	
	Method AddRaw(Raw:String)
		Enqueue(Raw)
	End Method
	
	Method AddArray(Source:TASTNode[], Separator:String)
		For Local I:Int = 0 Until Source.Length
			If I Then AddRaw(Separator)
			Source[I].GenerateCode(Self)
		Next
	End Method
	
	Method AddList(Source:TList, Separator:String)
		For Local Node:TASTNode = EachIn Source
			If Node <> Source.First() Then AddRaw(Separator)
			Node.GenerateCode(Self)
		Next
	End Method
	
	Method AddBracketed(Source:TASTNode)
		AddRaw("(")
		Source.GenerateCode(Self)
		AddRaw(")")
	End Method
	
	Method AddIndentedRaw(Raw:String)
		AddRaw(IndentString + Raw)
	End Method
	
	Method AddDeclaration(Source:TASTNode, Prefix:String = "")
		BeginStatement(Source)
		Source.GenerateDeclaration(Self, Prefix)
		EndStatement()
	End Method
	
	Method AddDeclarationList(Source:TList, Separator:String)
		For Local Node:TASTNode = EachIn Source
			If Node <> Source.First() Then AddRaw(Separator)
			Node.GenerateDeclaration(Self)
		Next
	End Method
	
	Method AddImplementation(Source:TASTNode, Prefix:String = "")
		BeginStatement(Source)
		Source.GenerateImplementation(Self, Prefix)
		EndStatement()
	End Method
	
	Method AddImplementationList(Source:TList, Separator:String)
		For Local Node:TASTNode = EachIn Source
			If Node <> Source.First() Then AddRaw(Separator)
			Node.GenerateImplementation(Self)
		Next
	End Method
	
	Method BeginScope()
		BeginInvisibleScope()
		
		AddRaw("{")
		AddNewline()
		
		IndentationLevel :+ 1
		IndentString :+ "~t"
	End Method
	
	Method EndScope()
		IndentationLevel :- 1
		IndentString = IndentString[.. IndentationLevel]
		
		AddIndentedRaw("}")
		
		EndInvisibleScope()
	End Method
	
	Method BeginInvisibleScope()	
		CurrentScope = New TDebugScope.Create(CurrentScope)
		CurrentScope.StartLine = CurrentLine
	End Method
	
	Method EndInvisibleScope()
		CurrentScope.EndLine = CurrentLine
		CurrentScope = CurrentScope.Parent
	End Method
	
	Method BeginStatement(Source:TASTNode)
		SetCurrent(Source)
		AddRaw(IndentString)
	End Method
	
	Method EndStatement()
		AddRaw(";")
		AddNewline()
	End Method
	
	Method BeginIncludeGuard(File:String)
		Local Guard:String = StripAll(File).ToUpper() + "_" + Hex(Rand(0, $FFFFFF))
		
		AddRaw("#ifndef " + Guard)
		AddNewline()
		AddRaw("#define " + Guard)
		AddNewline()
	End Method
	
	Method EndIncludeGuard()
		AddRaw("#endif")
		AddNewline()
	End Method
	
	Method GenerateCode:String()
		Local OutLength:Int
		
		For Local Pair:TCPPOutputPair = EachIn OutputQueue
			OutLength :+ Pair.Payload.Length
		Next
		
		Local OutBytes:Short[] = New Short[OutLength], Index:Int
		For Local Pair:TCPPOutputPair = EachIn OutputQueue
			For Local I:Int = 0 Until Pair.Payload.Length
				OutBytes[Index] = Pair.Payload[I]
				Index :+ 1
			Next
		Next
		
		Return String.FromShorts(OutBytes, OutLength)
	End Method
	
	Method GenerateDebugInfo:String()
		CurrentScope.EndLine = CurrentLine
		
		Return GenerateLineMapping() + "~n" + CurrentScope.ToString()
	End Method
	
	Method GenerateLineMapping:String()
		Local Lines:String[] = New String[CurrentLine + 1]
		
		For Local Pair:TCPPOutputPair = EachIn OutputQueue
			If Pair.Source Then ..
				Lines[Pair.Line] = Pair.Line + ":" + Pair.Source.StartLine
		Next
		
		Local Out:String
		For Local I:String = EachIn Lines
			If I Then Out :+ I + "~n"
		Next
		
		Return Out.Trim()
	End Method
	
	Method Enqueue(Payload:String)
		OutputQueue.AddLast(New TCPPOutputPair.Create(Payload, CurrentNode, CurrentLine))
	End Method
End Type

Type TCPPOutputPair
	Field Payload:String
	Field Source:TASTNode
	Field Line:Int
	
	Method Create:TCPPOutputPair(Payload:String, Source:TASTNode, Line:Int)
		Self.Payload = Payload
		Self.Source  = Source
		Self.Line    = Line
		
		Return Self
	End Method
End Type

Type TDebugVariable
	Field Variable:TVariableDeclNode
	Field Line:Int
	
	Method Create:TDebugVariable(Variable:TVariableDeclNode, Line:Int)
		Self.Variable = Variable
		Self.Line     = Line
		
		Return Self
	End Method
	
	Method ToString:String()
		Return Line + " " + Variable.Identifier + ":" + Variable.Datatype.ToString()
	End Method
End Type

Type TDebugScope
	Field StartLine:Int
	Field EndLine:Int
	
	Field Parent:TDebugScope
	Field Children:TList
	
	Method New()
		Children = New TList
	End Method
	
	Method AddChild(Child:Object)
		Children.AddLast(Child)
	End Method
	
	Method Create:TDebugScope(Parent:TDebugScope)
		Self.Parent = Parent
		
		If Parent Then Parent.AddChild(Self)
		
		Return Self
	End Method
	
	Method ToString:String()
		Local Out:String = StartLine + " {~n"
		
		For Local Child:Object = EachIn Children
			Out :+ "~t" + Child.ToString().Replace("~n", "~n~t") + "~n"
		Next
		
		Return Out + EndLine + " }"
	End Method
End Type