Type TModule
	Global ModuleFolder:String = "mod"
	Global AbsoluteModuleFolder:String = RealPath(ModuleFolder)
	
	Global LoadedModules:TList = New TList
	
	Field Dependencies:TDependencyNode
	Field ExportScope:TScope
	
	Field Root:TProgramNode
	
	Field Libraries:TList
	
	Field Identifier:String
	Field Path:String
	Field Scope:String
	
	Field IsModule:Int
	Field IsOutdated:Int
	Field CodePath:String
	
	Field Parent:TModule
	Field Children:TList
	
	Field ImportRequired:Int
	
	Field DoesCompile:Int
	Field Dependers:TList
	Field MissingSymbols:TList
	Field MissingModules:TList
	
	Method New()
		Children       = New TList
		Libraries      = New TList
		Dependencies   = New TDependencyNode
		Dependers      = New TList
		DoesCompile    = True
		MissingSymbols = New TList
		MissingModules = New TList
	End Method
	
	Method Load:TModule(Parent:TModule, Identifier:String)
		Self.Identifier = Identifier
		Self.Parent = Parent
		
		If Parent Then
			If Parent.Scope Then
				Scope = Parent.Scope + "." + Identifier
			Else
				Scope = Identifier
			EndIf
			
			Path  = Parent.Path  + "/" + Identifier
		Else
			Scope = ""
			Path  = ModuleFolder
		EndIf
		
		Local Dir:Int = ReadDir(Path)
		
		Local File:String = NextFile(Dir)
		While File
			If FileType(Path + "/" + File) = FILETYPE_DIR And File <> "." And File <> ".." And File[0] <> "."[0] Then ..
				Children.AddLast(New TModule.Load(Self, File))
			
			File = NextFile(Dir)
		Wend
		
		CloseDir(Dir)
		
		CodePath = RealPath(Path + "/" + Identifier + "." + TCompiler.FileExtension)
		If FileType(CodePath) Then
			IsModule = True
			Dependencies.Create(CodePath, Self)
		EndIf
		
		Return Self
	End Method
	
	Method LoadAST()
		IsOutdated = TCompiler.ModuleCompilationNecessary(CodePath, Root)
		
		If Not Root Then Root = TCompiler.LoadRootFromDep(TCompiler.GenerateModuleTargetPath(CodePath) + ".dep", CodePath)
		
		If Root Then
			Root.Dependencies = Dependencies
			Root.IsModule = True
			Libraries = Root.Libraries
		EndIf
	End Method
	
	Method Find:TModule(Path:String)
		Local Split:Int = Path.Find(".")
		If Split < 0 Then Split = Path.Length
		
		Local Ident:String = Path[.. Split].ToLower()
		
		If Ident = "*" Then Return Self
		
		For Local SubModule:TModule = EachIn Children
			If SubModule.Identifier.ToLower() = Ident Then
				If Split = Path.Length Then Return SubModule Else Return SubModule.Find(Path[Split + 1 ..])
			EndIf
		Next
	End Method
	
	Method ToList:TList(List:TList = Null)
		If Not List Then
			List = New TList
		Else
			List.AddLast(Scope)
		EndIf
		
		For Local Child:TModule = EachIn Children
			Child.ToList(List)
		Next
		
		Return List
	End Method
	
	Function ListModulesInScope:TList(Scope:String, List:TList = Null)
		If Not List Then List = New TList
		
		Local LoadRecursive:Int
		If Scope.EndsWith(".*") Then
			LoadRecursive = True
			Scope = Scope[.. Scope.Length - 2]
		EndIf
		
		Local CodePath:String = ScopeToCodePath(Scope)
		If FileType(CodePath) Then List.AddLast(CodePath)
		
		If LoadRecursive Then
			Local Folders:String[] = LoadDir(ScopeToFolderPath(Scope))
			For Local I:String = EachIn Folders
				If Not IgnoreFolder(I) Then ListModulesInScope(Scope + "." + I + ".*", List)
			Next
		EndIf
		
		Return List
	End Function
	
	Function ScopeToFolderPath:String(Scope:String)
		Return AbsoluteModuleFolder + "/" + Scope.Replace(".", "/")
	End Function
	
	Function ScopeToCodePath:String(Scope:String)
		Local Name:String = Scope
		If Scope.Find(".") >= 0 Then Name = ExtractExt(Scope)
		
		Return AbsoluteModuleFolder + "/" + Scope.Replace(".", "/") + "/" + Name + ".bi"
	End Function
End Type

Type TModuleScope Extends TScope
	Const MODE_INSERT :Int = 0
	Const MODE_MODULE :Int = 1
	Const MODE_PROGRAM:Int = 2
	
	Field Mode:Int
	
	Field ModuleList:TList
	
	Field CurrentModule:TModule
	
	Method CreateModuleScope:TModuleScope(Parent:TScope, ModuleList:TList)
		Self.ModuleList = ModuleList.Reversed()
		Self.Parent = Parent
		
		Mode = MODE_INSERT
		
		TASTNode.IgnoreSemanticErrors = True
		For Local M:TModule = EachIn ModuleList
			M.ExportScope = New TScope.Create(Self)
			
			If M.Root Then M.Root.CheckCorrectness(M.ExportScope)
		Next
		TASTNode.IgnoreSemanticErrors = False
		
		Mode = MODE_MODULE
		
		Return Self
	End Method
	
	Method FinishSelfCheck()
		Mode = MODE_PROGRAM
	End Method
	
	Method RemoveModule(M:TModule)
		ModuleList.Remove(M)
	End Method
	
	Method SetCurrent(M:TModule)
		CurrentModule = M
	End Method
	
	Method Insert:Int(Identifier:String, Value:TASTNode)
		Return True
	End Method
	
	Method Find:TASTNode(Identifier:String)
		If Mode = MODE_INSERT Then
			Return Null
		Else
			For Local M:TModule = EachIn ModuleList
				M.ExportScope.Parent = Null 'Avoid infinite recursion
				Local Node:TASTNode = M.ExportScope.Find(Identifier)
				M.ExportScope.Parent = Self
				
				If Node Then
					If Mode = MODE_MODULE Then M.Dependers.AddLast(CurrentModule)
					
					Return Node
				EndIf
			Next
			
			If Parent Then
				Local Node:TASTNode = Parent.Find(Identifier)
				
				If Node Then Return Node
			EndIf
			
			If Mode = MODE_MODULE Then
				CurrentModule.DoesCompile = False
				CurrentModule.MissingSymbols.AddLast(Identifier)
				
				Return Null
			Else
				Return Null
			EndIf
		EndIf
	End Method
End Type